@@ -31,20 +31,24 @@
 
 #define ALIGNED_STRING_SIZE(len) (((len) + sizeof (void*) + WORD_ALIGN_1) & WORD_ALIGN_MASK)
 
-#define PARROT_GC_WRITE_BARRIER(i, p) do { if (PObj_GC_need_write_barrier_TEST((p))) Parrot_gc_write_barrier((i), (p)); } while(0)
+#define PARROT_GC_WRITE_BARRIER(i, p) \
+    do { if (PObj_GC_need_write_barrier_TEST((p))) Parrot_gc_write_barrier((i), (p)); } while (0)
 
 typedef struct _Parrot_GC_Init_Args {
     void *stacktop;
     const char *system;
+    Parrot_Float4 nursery_size;
     Parrot_Int dynamic_threshold;
     Parrot_Int min_threshold;
+    Parrot_UInt numthreads;
+    Parrot_UInt debug_flags;
 } Parrot_GC_Init_Args;
 
 typedef enum _gc_sys_type_enum {
     MS,  /* mark and sweep */
     INF, /* infinite memory core */
-    MS2,
-    GMS
+    MS2, /* non-recursive mark and sweep */
+    GMS  /* generational mark & sweep */
 } gc_sys_type_enum;
 
 /* pool iteration */
@@ -67,14 +71,17 @@ typedef enum {
     GC_TRACE_SYSTEM_ONLY = 3
 } Parrot_gc_trace_type;
 
-typedef int (*pool_iter_fn)(PARROT_INTERP, struct Memory_Pools *, struct Fixed_Size_Pool *, int, void*);
-typedef void (*add_free_object_fn_type)(PARROT_INTERP, struct Memory_Pools *, struct Fixed_Size_Pool *, void *);
-typedef void * (*get_free_object_fn_type)(PARROT_INTERP, struct Memory_Pools *, struct Fixed_Size_Pool *);
-typedef void (*alloc_objects_fn_type)(PARROT_INTERP, struct Memory_Pools *, struct Fixed_Size_Pool *);
+typedef int (*pool_iter_fn)(PARROT_INTERP, struct Memory_Pools *,
+                struct Fixed_Size_Pool *, int, void*);
+typedef void (*add_free_object_fn_type)(PARROT_INTERP, struct Memory_Pools *,
+                struct Fixed_Size_Pool *, void *);
+typedef void * (*get_free_object_fn_type)(PARROT_INTERP, struct Memory_Pools *,
+                struct Fixed_Size_Pool *);
+typedef void (*alloc_objects_fn_type)(PARROT_INTERP, struct Memory_Pools *,
+                struct Fixed_Size_Pool *);
 typedef void (*gc_object_fn_type)(PARROT_INTERP, ARGMOD(struct Memory_Pools *),
                 ARGIN(struct Fixed_Size_Pool *), ARGMOD(PObj *));
 
-
 /* &gen_from_enum(interpinfo.pasm) prefix(INTERPINFO_) */
 
 typedef enum {
@@ -93,19 +100,32 @@ typedef enum {
     GC_LAZY_MARK_RUNS,
     EXTENDED_PMCS,
     CURRENT_RUNCORE,
+    PARROT_INTSIZE,
+    PARROT_FLOATSIZE,
+    PARROT_POINTERSIZE,
+    PARROT_INTMAX,
+    PARROT_INTMIN,
 
     /* interpinfo_p constants */
     CURRENT_CTX,
     CURRENT_SUB,
     CURRENT_CONT,
-    CURRENT_OBJECT,
     CURRENT_LEXPAD,
+    CURRENT_TASK,
 
     /* interpinfo_s constants */
     EXECUTABLE_FULLNAME,
     EXECUTABLE_BASENAME,
     RUNTIME_PREFIX,
-    GC_SYS_NAME
+    GC_SYS_NAME,
+    PARROT_OS,
+    PARROT_OS_VERSION,
+    PARROT_OS_VERSION_NUMBER,
+    CPU_ARCH,
+    CPU_TYPE,
+
+    /* additional gc constants */
+    MAX_GENERATIONS
 } Interpinfo_enum;
 
 /* &end_gen */
@@ -113,8 +133,10 @@ typedef enum {
 #define GC_trace_stack_FLAG    (UINTVAL)(1 << 1)   /* trace system areas and stack */
 #define GC_trace_normal_FLAG   (UINTVAL)(1 << 1)   /* the same */
 #define GC_lazy_FLAG           (UINTVAL)(1 << 2)   /* timely destruction run */
-#define GC_finish_FLAG         (UINTVAL)(1 << 3)   /* on Parrot exit: mark (almost) all PMCs dead and */
-#define GC_strings_cb_FLAG     (UINTVAL)(1 << 4)   /* Invoked from String GC during mem_alloc to sweep dead strings */
+#define GC_finish_FLAG         (UINTVAL)(1 << 3)   /* on Parrot exit: mark (almost)
+                                                      all PMCs dead and */
+#define GC_strings_cb_FLAG     (UINTVAL)(1 << 4)   /* Invoked from String GC during
+                                                      mem_alloc to sweep dead strings */
                                                    /* garbage collect. */
 
 /* HEADERIZER BEGIN: src/gc/api.c */
@@ -124,6 +146,10 @@ PARROT_EXPORT
 void Parrot_block_GC_mark(PARROT_INTERP)
         __attribute__nonnull__(1);
 
+PARROT_EXPORT
+void Parrot_block_GC_mark_locked(PARROT_INTERP)
+        __attribute__nonnull__(1);
+
 PARROT_EXPORT
 void Parrot_block_GC_sweep(PARROT_INTERP)
         __attribute__nonnull__(1);
@@ -138,7 +164,7 @@ int Parrot_gc_active_sized_buffers(PARROT_INTERP)
 
 PARROT_EXPORT
 void Parrot_gc_allocate_buffer_storage_aligned(PARROT_INTERP,
-    ARGOUT(Buffer *buffer),
+    ARGOUT(Parrot_Buffer *buffer),
     size_t size)
         __attribute__nonnull__(1)
         __attribute__nonnull__(2)
@@ -209,7 +235,7 @@ void Parrot_gc_finalize(PARROT_INTERP)
 
 PARROT_EXPORT
 void Parrot_gc_free_bufferlike_header(PARROT_INTERP,
-    ARGMOD(Buffer *obj),
+    ARGMOD(Parrot_Buffer *obj),
     size_t size)
         __attribute__nonnull__(1)
         __attribute__nonnull__(2)
@@ -224,15 +250,12 @@ void Parrot_gc_free_fixed_size_storage(PARROT_INTERP,
         FUNC_MODIFIES(*data);
 
 PARROT_EXPORT
-PARROT_CANNOT_RETURN_NULL
 void Parrot_gc_free_memory_chunk(PARROT_INTERP, ARGIN_NULLOK(void *data))
         __attribute__nonnull__(1);
 
 PARROT_EXPORT
-void Parrot_gc_free_pmc_attributes(PARROT_INTERP, ARGMOD(PMC *pmc))
-        __attribute__nonnull__(1)
-        __attribute__nonnull__(2)
-        FUNC_MODIFIES(*pmc);
+void Parrot_gc_free_pmc_attributes(PARROT_INTERP, ARGFREE(PMC *pmc))
+        __attribute__nonnull__(1);
 
 PARROT_EXPORT
 void Parrot_gc_free_pmc_header(PARROT_INTERP, ARGMOD(PMC *pmc))
@@ -280,6 +303,10 @@ void Parrot_gc_mark_STRING_alive_fun(PARROT_INTERP,
         __attribute__nonnull__(1)
         FUNC_MODIFIES(*obj);
 
+PARROT_EXPORT
+int Parrot_gc_max_generations(PARROT_INTERP)
+        __attribute__nonnull__(1);
+
 PARROT_EXPORT
 size_t Parrot_gc_mem_alloc_since_last_collect(PARROT_INTERP)
         __attribute__nonnull__(1);
@@ -287,7 +314,7 @@ size_t Parrot_gc_mem_alloc_since_last_collect(PARROT_INTERP)
 PARROT_EXPORT
 PARROT_CANNOT_RETURN_NULL
 PARROT_WARN_UNUSED_RESULT
-Buffer * Parrot_gc_new_bufferlike_header(PARROT_INTERP, size_t size)
+Parrot_Buffer * Parrot_gc_new_bufferlike_header(PARROT_INTERP, size_t size)
         __attribute__nonnull__(1);
 
 PARROT_EXPORT
@@ -310,7 +337,7 @@ void Parrot_gc_pmc_needs_early_collection(PARROT_INTERP, ARGMOD(PMC *pmc))
 
 PARROT_EXPORT
 void Parrot_gc_reallocate_buffer_storage(PARROT_INTERP,
-    ARGMOD(Buffer *buffer),
+    ARGMOD(Parrot_Buffer *buffer),
     size_t newsize)
         __attribute__nonnull__(1)
         __attribute__nonnull__(2)
@@ -381,12 +408,18 @@ PARROT_EXPORT
 void Parrot_unblock_GC_mark(PARROT_INTERP)
         __attribute__nonnull__(1);
 
+PARROT_EXPORT
+void Parrot_unblock_GC_mark_locked(PARROT_INTERP)
+        __attribute__nonnull__(1);
+
 PARROT_EXPORT
 void Parrot_unblock_GC_sweep(PARROT_INTERP)
         __attribute__nonnull__(1);
 
 #define ASSERT_ARGS_Parrot_block_GC_mark __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
        PARROT_ASSERT_ARG(interp))
+#define ASSERT_ARGS_Parrot_block_GC_mark_locked __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
+       PARROT_ASSERT_ARG(interp))
 #define ASSERT_ARGS_Parrot_block_GC_sweep __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
        PARROT_ASSERT_ARG(interp))
 #define ASSERT_ARGS_Parrot_gc_active_pmcs __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
@@ -443,8 +476,7 @@ void Parrot_unblock_GC_sweep(PARROT_INTERP)
 #define ASSERT_ARGS_Parrot_gc_free_memory_chunk __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
        PARROT_ASSERT_ARG(interp))
 #define ASSERT_ARGS_Parrot_gc_free_pmc_attributes __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
-       PARROT_ASSERT_ARG(interp) \
-    , PARROT_ASSERT_ARG(pmc))
+       PARROT_ASSERT_ARG(interp))
 #define ASSERT_ARGS_Parrot_gc_free_pmc_header __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
        PARROT_ASSERT_ARG(interp) \
     , PARROT_ASSERT_ARG(pmc))
@@ -469,6 +501,8 @@ void Parrot_unblock_GC_sweep(PARROT_INTERP)
 #define ASSERT_ARGS_Parrot_gc_mark_STRING_alive_fun \
      __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
        PARROT_ASSERT_ARG(interp))
+#define ASSERT_ARGS_Parrot_gc_max_generations __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
+       PARROT_ASSERT_ARG(interp))
 #define ASSERT_ARGS_Parrot_gc_mem_alloc_since_last_collect \
      __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
        PARROT_ASSERT_ARG(interp))
@@ -519,6 +553,8 @@ void Parrot_unblock_GC_sweep(PARROT_INTERP)
        PARROT_ASSERT_ARG(interp))
 #define ASSERT_ARGS_Parrot_unblock_GC_mark __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
        PARROT_ASSERT_ARG(interp))
+#define ASSERT_ARGS_Parrot_unblock_GC_mark_locked __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
+       PARROT_ASSERT_ARG(interp))
 #define ASSERT_ARGS_Parrot_unblock_GC_sweep __attribute__unused__ int _ASSERT_ARGS_CHECK = (\
        PARROT_ASSERT_ARG(interp))
 /* Don't modify between HEADERIZER BEGIN / HEADERIZER END.  Your changes will be lost. */
@@ -527,10 +563,23 @@ void Parrot_unblock_GC_sweep(PARROT_INTERP)
 # define Parrot_gc_mark_STRING_alive(interp, obj) Parrot_gc_mark_STRING_alive_fun((interp), (obj))
 
 #if defined(PARROT_IN_CORE)
+#ifdef THREAD_DEBUG
 #  define Parrot_gc_mark_PMC_alive(interp, obj) \
-      do if (!PMC_IS_NULL(obj)) Parrot_gc_mark_PMC_alive_fun((interp), (obj)); \
+      do if (!PMC_IS_NULL(obj) \
+          && (!PObj_is_shared_TEST(obj) || !Interp_flags_TEST((interp), PARROT_IS_THREAD))) { \
+            PARROT_ASSERT((obj)->orig_interp == (interp)); \
+            Parrot_gc_mark_PMC_alive_fun((interp), (obj)); \
+        } \
       while (0)
 #else
+#  define Parrot_gc_mark_PMC_alive(interp, obj) \
+      do if (!PMC_IS_NULL(obj) \
+          && (!PObj_is_shared_TEST(obj) || !Interp_flags_TEST((interp), PARROT_IS_THREAD))) { \
+          Parrot_gc_mark_PMC_alive_fun((interp), (obj)); \
+        } \
+      while (0)
+#endif
+#else
 #  define Parrot_gc_mark_PMC_alive(interp, obj) Parrot_gc_mark_PMC_alive_fun((interp), (obj))
 #endif
 
